# coding: utf-8
"""
    Simple condition parser
    Copyright (c) 2008 Friedrich Weber

    Based on the simpleBool.py pyparsing example by
    Copyright (c) 2003-2008  Paul T. McGuire

    Permission is hereby granted, free of charge, to any person obtaining
    a copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to
    permit persons to whom the Software is furnished to do so, subject to
    the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import pyparsing as pyp

class ParseException(Exception):
    def __init__(self, message):
        """ little helper exception for parsing errors """
        Exception.__init__(self)
        self.message = message

    def __str__(self):
        """ return my message """
        return self.message

class ConditionHandler():
    """ class for handling conditions,
        contains pyparsing-internal classes """
    class _BoolOperand(object):
        def __init__(self,t):
            self.args = t[0][0::2]

        def __repr__(self):
            sep = " %s " % self.reprsymbol
            return "(" + sep.join(map(str,self.args)) + ")"

    class _BoolAnd(_BoolOperand):
        reprsymbol = "&"
        def __nonzero__(self):
            for a in self.args:
                if isinstance(a,basestring):
                    v = eval(a)
                else:
                    v = bool(a)
                if not v:
                    return False
            return True

    class _BoolOr(_BoolOperand):
        reprsymbol = '|'
        def __nonzero__(self):
            for a in self.args:
                if isinstance(a,basestring):
                    v = eval(a)
                else:
                    v = bool(a)
                if v:
                    return True
            return False

    class _BoolNot(_BoolOperand):
        def __init__(self,t):
            self.arg = t[0][1]

        def __repr__(self):
            return "~" + str(self.arg)

        def __nonzero__(self):
            if isinstance(self.arg,basestring):
                v = eval(self.arg)
            else:
                v = bool(self.arg)
            return not v

    """ --- main --- """
    def __init__(self):
        """ initialize """
        self.functions = {}
        # set pyparsing expressions
        self.func_expression = pyp.Word(pyp.alphanums+'_') + \
                '('+ \
                pyp.OneOrMore(pyp.QuotedString('"', '\\') + pyp.Optional(" ")) \
                + ')'
        self.func_expression.setParseAction(self._call_function)

        bool_operand = pyp.oneOf("True False")
        self.bool_expr = pyp.operatorPrecedence( bool_operand,
            [
            ("not", 1, pyp.opAssoc.RIGHT, self._BoolNot),
            ("or",  2, pyp.opAssoc.LEFT,  self._BoolOr),
            ("and", 2, pyp.opAssoc.LEFT,  self._BoolAnd)
            ])

    def check(self, s):
        """ parse the condition `s` and return boolean """
        # is it empty? => return True
        if s == '':
            return True
        else:
            # call all functions and create a boolean expression
            transformed = self.func_expression.transformString(s)
            # parse the boolean expression
            ret = self.bool_expr.parseString(transformed)[0]
            if type(ret) == type(''):
                return ret == 'True'
            else:
                return bool(ret)

    def add_function(self, name, func):
        """ add a function. Note: all arguments will be passed as strings.
            :param name: name of the function
            :param func: Function
        """
        self.functions[name.lower()] = func

    def _call_function(self, string, position, tokens):
        """ internal pyparsing callback that calls the function.
            :param string: The complete condition string
            :param position: The position of the token start
            :param tokens: The tokens. Should be
                    ['FunctionName', '(', 'Argument #1', 'Argument #2', ')']
        """
        # split the tokens
        try:
            func_name = tokens[0].lower()
            arguments = tokens[2:-1]
        except Exception, e:
            raise ParseException(str(e))
            return ['False']
        # is there a matching function?
        if func_name not in self.functions:
            raise ParseException(_("No function named '%s'") % func_name)
            return ['False']
        # call the function and return its result
        result = self.functions[func_name](*arguments)
        # test result
        if str(result).lower() not in ['true', 'false']:
            raise ParseException(_("Invalid function result: '%s'") \
                                                        % repr(result))
            return ['False']
        # done!
        return [str(result)]

if __name__ == "__main__":
    ch = ConditionHandler()
    # add callbacks. Don't have to be lambdas, of course.
    ch.add_function('is_foo', lambda x: x.lower() == 'foo')
    ch.add_function('is_bar', lambda x: x.lower() == 'bar')
    ch.add_function('is_foobar', lambda x, y: x == 'foo' and y == 'bar')
    test = 'is_foo("Foo") or is_bar("Foo")'
    print ch.check(test)
