#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    GESTIKK
    Version 0.6.1

    Gestikk, a mouse gesture recognition application
    Copyright (C) 2007-2008:
    Friedrich 'fred.reichbier' Weber (2007-10-08) and
    Jonas 'Dauerbaustelle' Haag (2008)

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""

import os
import thread
import subprocess
import signal
import gtk
import gobject
import pynotify
if not pynotify.init("gestikk"):
    print "Libnotify daemon error!"

import osd
from gestikk.tools import *

_ = init_gettext().gettext

_check_bin = lambda a, b: a & b == b
""" checks whether a & b == b """

class StatusIcon:
    """ handle the status icon in the GNOME tray """
    def __init__(self):
        """ initialize """
        # initialize the gtk.StatusIcon
        icon = gtk.StatusIcon()
        # save
        self.icon = icon
        # set the icon
        icon.set_from_file(get_file('gestikk.xpm', 'icon'))
        # set visible
        icon.set_visible(True)
        # - set menu
        menu = gtk.Menu()
        # "Configuration"
        menu_item = gtk.ImageMenuItem(gtk.STOCK_PREFERENCES)
        menu_item.connect('activate', self.evt_preferences)
        menu.append(menu_item)
        # "About"
        menu_item = gtk.ImageMenuItem(gtk.STOCK_ABOUT)
        menu_item.connect('activate', self.evt_about)
        menu.append(menu_item)
        # Separator
        menu_item = gtk.SeparatorMenuItem()
        menu.append(menu_item)
        # "Close"
        menu_item = gtk.ImageMenuItem(gtk.STOCK_CLOSE)
        menu_item.connect('activate', self.evt_close)
        menu.append(menu_item)
        # - set icon signals
        icon.connect('popup-menu', self.evt_menu, menu)

    def evt_close(self, evt):
        """ close """
        gtk.main_quit()

    def evt_about(self, evt):
        """ show about dialog """
        def close(widget, res):
            if res == gtk.RESPONSE_CANCEL:
                widget.hide()
        about_dialog = get_about_dialog()
        about_dialog.connect("response", close) # Todo?
        about_dialog.run()

    def evt_menu(self, widget, button, time, data = None):
        """ show popup menu """
        if button == 3 and data:
            data.show_all()
            data.popup(None, None, None, 3, time)

    def evt_preferences(self, evt):
        """ launch preferences frame """
        try:
            subprocess.Popen([get_file('gestikk.py', 'executable'), '-c', '-q'])
        except Exception:
            subprocess.Popen([get_file('gestikk', 'executable'), '-c', '-q'])


class Notifier:
    """ class for notifying the user when a gesture started """
    def __init__(self, logic, statusicon):
        self.logic = logic
        self.statusicon = statusicon

    def notify(self, s, gesture):
        """ Notify and display `s`
            :type s: String
        """
        notify = pynotify.Notification(_("Notification"), s.replace('<', '&lt;').replace('>', '&gt;'))
        screen, rect, orient = self.statusicon.icon.get_geometry()
        notify.set_hint("x", rect.x + (rect.width / 2))
        notify.set_hint("y", rect.y + (rect.height / 2))
        notify.set_timeout(3000) # timeout in ms
        color = self.logic.config_get_option('OSD', 'color', '#0000FF')
        pb = pixmap_to_pixbuf(gesture_to_pixmap(gesture2py(gesture), 40, color), (255, 255, 255))
        notify.set_icon_from_pixbuf(pb)
        notify.show()

class Listener:
    """ class for listening the mouse positions """
    def __init__(self, logic, notifier, debugger):
        """
            :param logic: Logic() instance
            :param notifier: Notifier() instance
        """
        self.debugger = debugger
        self.logic = logic
        self.notifier = notifier
        self.root_window = gtk.gdk.get_default_root_window() # todo
        self.gesture_positions = []
        self.running = True
        self.x_win = None
        # coderedundanz mit sig_reload_gestures - TODO?
        self.notify_method = self.logic.config_get_option('Notification', 'type')
        self.use_gesture_osd = self.logic.config_get_option('OSD', 'enabled', '0') == '1'
        self.create_osd_window()

    def create_osd_window(self):
        """ (re-)create the osd window """
        if self.x_win:
            self.x_win.destroy()
        if self.use_gesture_osd:
            self.x_win = osd.Window(self.logic.config_get_option('OSD', 'color', 'red'), int(self.logic.config_get_option('OSD', 'line_width', '3')))

    def stop(self):
        """ stop the gtk thread """
        self.running = False

    def start(self):
        """ run the thread """
        gobject.timeout_add(100, self._listen)

    def osd_paint(self, x, y):
        """ paint the gesture in-place """
        self.x_win.paint_gesture_path(x, y)

    def osd_clear(self):
        """ clear screen """
        self.x_win.clear()

    def notify_string(self, gesture):
        return "Starting action with gesture "+gesture

    def _listen(self):
        """ Internal listener. Called by gobject.timeout_* stuff """
        # get x, y and mouse status (clicked etc.)
        x, y, state = self.root_window.get_pointer()
        # re-arm timer
        # and now look up the gestures
        # get status of the modifier / mouse
        mstate = _check_bin(state, \
                 {'rightdown': gtk.gdk.BUTTON3_MASK,
                  'leftdown': gtk.gdk.BUTTON1_MASK,
                  'middledown': gtk.gdk.BUTTON2_MASK,
                  'ctrldown': gtk.gdk.CONTROL_MASK,
                  'altdown': gtk.gdk.MOD1_MASK,
                  'shiftdown': gtk.gdk.SHIFT_MASK
                  }.get(self.logic.current_switcher)
                  ) # a not-so-nice workaround to avoid unnecessary function calls
        if mstate: # Gesture done
            self.gesture_positions.append((x, y))
            # and paint
            if self.use_gesture_osd:
                self.osd_paint(x, y)
        else: # button no longer pressed
            if self.gesture_positions: # end gesture
                # remove gesture
                if self.use_gesture_osd:
                    self.osd_clear()
                # calculate path
                gesture = []
                tolerance = int(self.logic.config_get_option('Preferences', 'tolerance', '25'))
                for gid in range(1, len(self.gesture_positions)): # Let first out
                    gesture.append(calc_direction(self.gesture_positions[gid-1], \
                                                self.gesture_positions[gid], tolerance))
                # Remove zeros
                while (0,0) in gesture:
                    gesture.remove((0,0))
                # Delete duplicates
                gesture = remove_duplicates(gesture)
                self.gesture_positions = []
                # Is an action assigned to the gesture?
                gesture_repr = repr(gesture).replace(' ', '')[1:-1] # without spaces and brackets
                #print self.debugger.debug('Gesture: %s' % gesture_repr)
                # condition check
#                if gesture_repr in self.logic.gestures:
                self.debugger.debug('recognized: '+gesture_repr)
                if gesture_repr in self.logic.gestures:
                    # has one / multiple actions
                    if self.logic.config_get_option('Preferences', 'global condition', ''): # should we do this on an other way? TODO, maybe
                        self.debugger.debug('checking global condition ...')
                        if not condition_handler.check(self.logic.config_get_option('Preferences', 'global condition', '')):
                            self.debugger.debug('global condition isn\'t fulfilled, abort')
                            return self.running
                        self.debugger.debug('fine.')
                    for action in self.logic.gestures[gesture_repr]:
                        self.debugger.debug('Condition: %s' % str(action.check_condition()))
                        if action.check_condition():
                            action.do()
                            self.debugger.debug(self.notify_string(gesture_repr))
                            if self.notify_method == 'balloon' or self.notify_method == 'osd':
                                self.notifier.notify(action.get_formatted(), action.gesture)
                            if self.notify_method == 'osd':
                                gobject.timeout_add(3000, self._osd_clear) # show it 2 sec
        return self.running

    def _osd_clear(self):
        """ clear the osd notifier """
        if isinstance(self.notifier, osd.OSDNotifier):
            self.notifier.clear()

    def sig_reload_gestures(self, signum, frame):
        """ we should reload the gestures file - we received a SIGUSR1 signal,
            probably from the gestikk configuration frame. """
        self.debugger.debug("Reloading configuration.")
        self.logic.reload_gestures()
        self.logic.reload_config()
        # reload x_win
        self.use_gesture_osd = self.logic.config_get_option('OSD', 'enabled', '0') == '1'
        if self.use_gesture_osd:
            self.create_osd_window()
        self.notify_method = self.logic.config_get_option('Notification', 'type')
        if isinstance(self.notifier, osd.OSDNotifier):
            self.notifier.destroy()
        if self.notify_method == 'balloon':
            self.notifier = Notifier(self.logic, self.logic.status_icon)
        elif self.notify_method == 'osd':
            notify_colorstr = self.logic.config_get_option('Notification', 'color', 'red')
            notify_font = self.logic.config_get_option('Notification', 'font', 'Sans')
            notify_fontsize = self.logic.config_get_option('Notification' , 'font_size', '20')
            self.notifier = osd.OSDNotifier(notify_colorstr, notify_font, notify_fontsize)
        else:
            self.notifier = None

def main(debug):
    """
        :param debug: Debug mode on?
    """
    debugger = Debugger(debug)
    # create Logic
    logic = Logic()
    status = StatusIcon()
    logic.status_icon = status # Not clean... TODO?
    # get notifiers
    notify_m = logic.config_get_option('Notification', 'type')
    if notify_m == 'balloon':
        debugger.debug("Using balloontip notification")
        notifier = Notifier(logic, status)
        osd_obj = False
    elif notify_m == 'osd':
        debugger.debug("Using OSD notification")
        notify_colorstr = logic.config_get_option('Notification', 'color', 'red')
        notify_font = logic.config_get_option('Notification', 'font', 'Sans')
        notify_fontsize = logic.config_get_option('Notification' , 'font_size', '20')
        notifier = osd.OSDNotifier(notify_colorstr, notify_font, notify_fontsize)
    else:
        notifier = None
        osd_obj = False
    listener = Listener(logic, notifier, debugger)
    listener.start()
    signal.signal(signal.SIGUSR1, listener.sig_reload_gestures)
    # add a quit handler
    gtk.quit_add(gtk.main_level(), listener.stop)
    # start the main loop
    gtk.main()

if __name__ == '__main__':
    main(True)
